/* 
 * $TSUKUBA_Release: Omni OpenMP/XcalableMP Compiler Software Version 0.5.1 $
 * $TSUKUBA_Copyright:
 *  Copyright (C) 2010 University of Tsukuba
 *  
 *  This software is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License version
 *  2.1 published by the Free Software Foundation.
 *  
 *  Please check the Copyright and License information in the files named
 *  COPYRIGHT and LICENSE under the top  directory of the Omni Compiler
 *  Software release kit.
 *  
 *  * The specification of XcalableMP has been designed by the XcalableMP
 *    Specification Working Group (http://www.xcalablemp.org/).
 *  
 *  * The development of this software is supported by "Seamless and
 *    Highly-productive Parallel Programming Environment for
 *    High-performance computing" project funded by Ministry of Education,
 *    Culture, Sports, Science and Technology, Japan.
 *  $
 */
package xcodeml.c.decompile;

import java.util.Set;

import xcodeml.XmObj;
import xcodeml.c.binding.XcScanningVisitor;
import xcodeml.c.binding.gen.*;

/**
 * Visitor search a XmObj tree for variables reference.
 */
public class XcVarVisitor extends XcScanningVisitor
{
    private Set<String> _variables = null;

    /**
     * Checks if a XmObj tree contains variable references.
     * 
     * @param visitable a XmObj tree.
     * @return true if the tree contains variable references.
     */
    public static boolean checkContainVar(XmObj visitable)
    {
        XcVarVisitor visitor = new XcVarVisitor();

        return ((IRVisitable)visitable).enter(visitor);
    }

    /**
     * Checks if a XmObj tree contains variable references.
     * 
     * @param visitable a XmObj tree.
     * @return true if the tree contains variable references.
     */
    public static boolean checkContainVar(IRVisitable visitable)
    {
        XcVarVisitor visitor = new XcVarVisitor();

        return visitable.enter(visitor);
    }

    /**
     * Creates a XcVarVisitor.
     */
    public XcVarVisitor()
    {
    }

    /**
     * Creates a XcVarVisitor.
     * 
     * @param variables a set of variable's name a visitor use.
     */
    public XcVarVisitor(Set<String> variables)
    {
        _variables = variables;
    }

    @Override
    public boolean enter(XbcVar visitable)
    {
        if(_variables != null)
            _variables.add(visitable.getContent());

        return true;
    }

    @Override
    public boolean enter(XbcVarAddr visitable)
    {
        if(_variables != null)
            _variables.add(visitable.getContent());

        return true;
    }

    @Override
    public boolean enter(XbcFuncAddr visitable)
    {
        if(_variables != null)
            _variables.add(visitable.getContent());

        return true;
    }

//     @Override
//     public boolean enter(XbcArrayRef visitable)
//     {
//         if(_variables != null)
//             _variables.add(visitable.getContent());

//         return true;
//     }

    @Override
    public boolean enter(XbcMemberRef visitable)
    {
        if(_variables != null)
            _enterChildren(visitable);

        return true;
    }

    @Override
    public boolean enter(XbcMemberAddr visitable)
    {
        _enterChildren(visitable);

        return true;
    }

    @Override
    public boolean enter(XbcMemberArrayRef visitable)
    {
        _enterChildren(visitable);

        return true;
    }

    @Override
    public boolean enter(XbcMemberArrayAddr visitable)
    {
        _enterChildren(visitable);

        return true;
    }
}
