/* 
 * $TSUKUBA_Release: Omni OpenMP/XcalableMP Compiler Software Version 0.5.1 $
 * $TSUKUBA_Copyright:
 *  Copyright (C) 2010 University of Tsukuba
 *  
 *  This software is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License version
 *  2.1 published by the Free Software Foundation.
 *  
 *  Please check the Copyright and License information in the files named
 *  COPYRIGHT and LICENSE under the top  directory of the Omni Compiler
 *  Software release kit.
 *  
 *  * The specification of XcalableMP has been designed by the XcalableMP
 *    Specification Working Group (http://www.xcalablemp.org/).
 *  
 *  * The development of this software is supported by "Seamless and
 *    Highly-productive Parallel Programming Environment for
 *    High-performance computing" project funded by Ministry of Education,
 *    Culture, Sports, Science and Technology, Japan.
 *  $
 */
package xcodeml.f.decompile;

import java.util.Deque;
import java.util.LinkedList;

import xcodeml.binding.IRNode;

/**
 * Decompile utility.
 */
class XfUtil
{
    public static void errorPrint(String message)
    {
        System.err.println(message);
    }

    /**
     * Convert Relaxer generated class name into an element name.
     * @param node Instance of IRNode
     * @return Element name.
     */
    public static String getElementName(IRNode node)
    {
        String nodeName = node.getClass().getSimpleName();
        if (nodeName.startsWith("Xbf") == false) {
            return nodeName;
        }
        return nodeName.substring(3);
    }

    public static String[] getNodePath(IRNode node)
    {
        Deque<String> list = new LinkedList<String>();
        list.addFirst(getElementName(node));
        node = node.rGetParentRNode();
        while (node != null) {
            list.addFirst(getElementName(node));
            node = node.rGetParentRNode();
        }
        return list.toArray(new String[0]);
    }

    public static String formatError(IRNode errorNode, XfError errorCode, Object... args)
    {
        StringBuilder sb = new StringBuilder();
        sb.append("Error(");
        sb.append(errorCode.ordinal());
        sb.append("): ");
        sb.append(errorCode.format(args));

        if (errorNode == null) {
            return sb.toString();
        }

        sb.append("\n<Element Tree>\n");

        String[] nodeArray = getNodePath(errorNode);
        if (nodeArray.length > 0) {
            sb.append(nodeArray[0]);
            sb.append("\n");
        }

        for (int i = 1; i < nodeArray.length; i++) {
            for (int j = 0; j < i; j++) {
                sb.append("  ");
            }
            sb.append("+");
            sb.append(nodeArray[i]);
            sb.append("\n");
        }
        return sb.toString();
    }

    public static boolean isNullOrEmpty(String s)
    {
        return ((s == null) || (s.isEmpty() != false));
    }
}
